#
# Vlatko Primorac, 1998, 1999
#

use strict;
use Tk;
use Tk::Dialog;
use Tk::LabEntry;
use Getopt::Std;   # single character options

my %args;          # command line options

getopts("hvpilr:c:d:a:n:ts:", \%args);
my $verbose = $args{v};            # verbose
my $scoreFile = $args{s};          # score file name
my $privateColormap = $args{p};    # private colormap
my $iconify = $args{i};            # start iconified 

#
# define the size of the board
#
my $boardSizeX = 9;
if( defined $args{c} ) { $boardSizeX = $args{c} };
my $boardSizeY = 9;
if( defined $args{r} ) { $boardSizeY = $args{r} };

#
# define number of kugels to add and delete
#
my $addKugel  = 3 ;
if( defined $args{a} ) { $addKugel = $args{a} };
my $delKugel  = 4 ;  # current one doesn't count
if( defined $args{d} ) { $delKugel = $args{d} - 1 };
my $setOf     = 7 ;  # number of different kugels to use
if( defined $args{n} ) { $setOf = $args{n} };
my $allowSkip = 0;
if( $args{t} ) { $allowSkip = 1 };
my $landscape = 0;
if( $args{l} ) { $landscape = 1 };  # landscape layout

if( $verbose ) {
    foreach my $arg (sort keys %args){
        print "Option $arg = $args{$arg}\n";
    }
}

#
# print command line help
#
if( $args{h} ) {
    print "\n";
    print "$0 [-p] [-r#] [-c#] [-a#] [-d#] [-t] [-s scorefile] [-v]\n";
    print "  -a #            add by #\n";
    print "  -c #            number of columns\n";
    print "  -d #            delete by #\n";
    print "  -h              print this message\n";
    print "  -i              start iconified\n";
    print "  -l              landscape layout\n";
    print "  -n #            use set of # kugels\n";
    print "  -p              use private colormap\n";
    print "  -r #            number of rows\n";
    print "  -s scorefile    specify alternate scorefile\n";
    print "  -t              allow skip through\n";
    print "  -v              verbose mode\n";
    print "\n";
    print 'This is version 2.0 of the perl/tk based  game I call "kugel",'."\n";
    print 'you can  call it whatever you  please.  The game  is played by'."\n";
    print 'moving  objects  on  the board  and putting  them  into  rows,'."\n";
    print 'columns, diagonals.  Each time you create  a line of  the same'."\n";
    print 'objects  ("kugels", balls...), the score  will  increment with'."\n";
    print 'the  number  of objects  you  accomplished to line up.  In the'."\n";
    print 'same time,  the lined up "kugels"  will  disappear, so you can'."\n";
    print 'continue the game with more maneuver  space on the board.  Any'."\n";
    print 'time  you  fail to create  a line of "kugels",  new ones  will'."\n";
    print 'appear on the board. Once the board is full, the game is over.'."\n";
    print "\n";
    print "The latest and greatest version of the game is available at:\n";
    print "http://www.sunsite.unc.edu/pub/Linux/games/strategy/\n";
    print "\n";
    print "Have fun. :)\n\nVlatko Primorac, 1998, 1999\nvlp\@geocities.com\n\n";
    exit 0;
}

#
# check for versions of Tk module and Perl itself, warn user if modules are old
#
my $tkVersion = $Tk::VERSION;

if( $verbose ) { print "Using Tk version: $tkVersion\n"; }
if( $verbose ) { print "Using Perl version: $]\n"; }

if( $] < 5.005 ) { 
    warn "This program has been tested with Perl version 5.00502\n".
         "You might have problems trying to use it with older version\n";
}
if( $tkVersion < 800.012 ) { 
    warn "This program has been tested with Tk module release 800.012\n".
         "You might have problems trying to use it with older version\n";
}

if( $verbose ) { print "Starting $0 ...\n" ; }

#
# create the main window
#
my $mainFrame;
if($privateColormap) {
    if( $verbose ) { print "Using private colormap.\n" ; }
    $mainFrame = MainWindow->new(-title => "Kugel", -colormap => "new");
} 
else {
    if( $verbose ) { print "Using default colormap.\n" ; }
    $mainFrame = MainWindow->new(-title => "Kugel"); 
}

if( $iconify ) {
    $mainFrame->iconify();  # start iconified
}

my $board = $mainFrame->Frame(-relief => 'raised', -borderwidth => 2);

#
# define pictures to use
#

my $imagesDir;

# search for "path/prgname" or "path\prgname.pl" or "path/prgname.pl"...
if( $0 =~ /(.*)[\\\/][|\\\/]*/ ) { $imagesDir = "$1" ; }
else { $imagesDir = "." }

if( ! $scoreFile ) { $scoreFile = "$imagesDir/scorefile"; }  # high score file

$imagesDir = "$imagesDir/images";

if( $verbose ) { print "Images directory: $imagesDir\n" ; }

opendir(DIR, $imagesDir)  || die "could not open $imagesDir : $!" ;
my @dir = readdir(DIR);
closedir(DIR);

@dir = sort @dir;
shift @dir; # get rid of .
shift @dir; # get rid of .

foreach my $dir (@dir) { $dir = $imagesDir."/".$dir; }

my %photo;

foreach my $dir (@dir) {
    opendir(DIR, $dir)  || die "could not open $dir : $!" ;
    my @imageFiles = readdir(DIR);
    closedir(DIR);

    @imageFiles = sort @imageFiles;

    shift @imageFiles;  # get rid of .
    shift @imageFiles;  # get rid of ..

    foreach my $imageFile (@imageFiles) {
        if( $verbose ) { print "Dir: $dir, Image: $imageFile\n" ; }

        #
        # create anonymous array and assign to it's refference  :)
        #
        push @{$photo{$dir}}, $board->Photo(-format => "gif",
                                            # -height => 50,
                                            # -width =>  50,
                                            -file => "$dir/$imageFile");}
}

# my @currentSet = @{$photo{'./images/set2'}} ;

my @currentSet;

if( defined $dir[1] ) {                  # if 2nd set exist, choose it
    @currentSet = @{$photo{ $dir[1] }} ;
} elsif ( defined $dir[0] ) {            # otherwise, take the 1st set
    @currentSet = @{$photo{ $dir[0] }} ;
} else {                                          # or die (no images found)
    die "no images found\n"  ;
}

if( $setOf > $#currentSet ) {
    $setOf = $#currentSet;
    if($verbose) { print "currentSet reduced to $#currentSet\n"; }
}

#
# color scheme
#
my $ApplPalette = 'goldenrod1';
$mainFrame->setPalette($ApplPalette);  # application color scheme

my @ApplColors = qw( Gray Gold Orange IndianRed Green SkyBlue);

my $bigFont     = $mainFrame->fontCreate(-family => 'courier', 
                                         -weight => 'bold', 
                                         -size => '18');

my $scoreFont   = $mainFrame->fontCreate(-family => 'courier', 
                                         -weight => 'bold', 
                                         -size => '24' );

my $helpFont   = $mainFrame->fontCreate(-family => 'courier', 
                                        -weight => 'bold', 
                                        -size => '18' );

my $nextFrame = $mainFrame->Frame();
my @topKugel;

#
# the score
#
my $score=0;

my $filled=0;  #number of kugels on the screen

my @square;
my @row;
my $didSkip=0;       # did user ever use skip?
my $selected=undef;  # previously selected position
my @scores=undef;
my $highestScore=0;
my $champion="";

my $showShortcuts = 1;  # show bottom shotrtcuts by default
my $showInfo      = 1;  # show "semaphoreFrame"
my $showNext      = 1;  # show next set of kugels

&get_scores;

my $semaphoreFrame = $mainFrame->Frame();
#
# info frame
#
my $infoFrame = $semaphoreFrame->Frame(-relief => 'raised', -borderwidth => 2);
#
# i won't display this label, but i'll leave it here,
# it might be used it in some other version
#
my $numEmptyLabel=$infoFrame->Button(
    -text => "Empty: ".(sprintf '%02d',($boardSizeX*$boardSizeY-$filled))." ("
             .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
                              /($boardSizeX*$boardSizeY)))
             ."\%)",
    -command=>[\&help],
    -font => $scoreFont);

my $setOfLabel = $infoFrame->Button(
    -text => "Set Of $setOf ",
    -command=>[\&set_dimensions],
    -font => $scoreFont);

my $addLabel = $infoFrame->Button(
    -text => "Add by $addKugel ",
    -command=>[\&set_dimensions],
    -font => $scoreFont);

my $delLabel = $infoFrame->Button(
    -text => "Delete by ".($delKugel+1) ,
    -command=>[\&set_dimensions],
    -font => $scoreFont);

#
# percentage of squares filled
#
my $numFilledLabel=$infoFrame->Button(
    -text => (sprintf '%02d',
             int(100*$filled/($boardSizeX*$boardSizeY)))."\% Filled",
    -command=>[\&help],
    -font => $scoreFont);

#
# score frame
#
my $scoreFrame = $semaphoreFrame->Frame(-relief => 'raised', -borderwidth => 2);
my $scoreLabel = $scoreFrame->Button(
    -text => "Score: ".(sprintf '%04d',$score),
    -command => [\&show_hall_of_fame],
    -font => $scoreFont);
my $highScoreLabel = $scoreFrame->Button(
    -text => "$champion: ".(sprintf '%04d',$highestScore),
    -command => [\&show_hall_of_fame],
    -font => $scoreFont);

#
# button menu shortcuts, down
#
my $bottomFrame = $mainFrame->Frame();

#
# color buttons, part of bottomFrame
#
my $colors = $bottomFrame->Frame();
foreach my $color (@ApplColors) {
    $colors->Button(
        -text   => $color,
        -pady => '0',
        -padx => '0',
        -background => $color,
        -activebackground => $color,
        -activeforeground => 'White',
        -command => sub { $mainFrame->setPalette($color); }
    )->pack(-side => 'left', -fill => 'both', -expand => 'no');
}
$colors->pack(-side => 'left', -fill => 'both', -expand => 'no');

my $restart = $bottomFrame->Button( -text => "restart",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&game_restart],);
my $quit = $bottomFrame->Button( -text => "quit program",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => sub{exit},);
my $kugelSets = $bottomFrame->Button( -text => "Kugel Sets",
                                      -pady => '0',
                                      -padx => '0',
                                      -command => [\&kugel_choice],);
my $dimensions = $bottomFrame->Button( -text => "Dimensions",
                                       -pady => '0',
                                       -padx => '0',
                                       -command => [\&set_dimensions],);
my $demo = $bottomFrame->Button( -text => "Demo",
                                 -pady => '0',
                                 -padx => '0',
                                 -command => [\&demo],);
$kugelSets->pack( -side => 'left', -fill => 'x', -expand => 'no');
$dimensions->pack( -side => 'left', -fill => 'x', -expand => 'no');
$demo->pack( -side => 'left', -fill => 'x', -expand => 'no');

$restart->pack( -side => 'left', -fill => 'x', -expand => 'no');
$quit->pack( -side => 'left', -fill => 'x', -expand => 'no');

#
# menu
#
my $menuFrame = $mainFrame->Frame(-relief => 'raised', -borderwidth => 2);

#
# file menu
#
my $fileMenu = $menuFrame->Menubutton(-text => 'File', -underline => 0);
$fileMenu->command(-label => 'Hall Of Fame', -command => [\&show_hall_of_fame]);
$fileMenu->command(-label => 'Restart', -command => [\&game_restart] );
$fileMenu->command(-label => 'Quit', -command => sub{exit} );
$fileMenu->pack(-side => 'left');

#
# options menu
#
my $optionsMenu = $menuFrame->Menubutton(-text => 'Options', -underline => 0);
$optionsMenu->command(-label => 'Kugel Choice', -command=>[\&kugel_choice]);
$optionsMenu->command(-label => 'Dimensions', -command=>[\&set_dimensions]);
$optionsMenu->checkbutton(
    -label    => 'Allow skip through',
    -variable => \$allowSkip,
    -command  => sub { 
                     &get_scores; 
                     if($verbose) { print "allowSkip = $allowSkip\n"; }
                 } 
);

$optionsMenu->pack(-side => 'left');

#
# view menu
#
my $viewMenu = $menuFrame->Menubutton(-text => 'View', -underline => 0);
$viewMenu->checkbutton(
    -label => 'landscape',
    -variable => \$landscape,
    -command => sub {
                    if($verbose) { print "landscape = $landscape\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Bottom Shortcuts',
    -variable => \$showShortcuts,
    -command => sub {
                    if($verbose) { print "showShortcuts = $showShortcuts\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Score Frame',
    -variable => \$showInfo,
    -command => sub {
                    if($verbose) { print "showInfo = $showInfo\n"; }
                    &re_pack;
                }
);
$viewMenu->checkbutton(
    -label => 'Next Kugel',
    -variable => \$showNext,
    command => sub {
                    if($verbose) { print "showNext = $showNext\n"; }
                    &re_pack;
                }
);
$viewMenu->pack(-side => 'left');

#
# colors menu
#
my $colorsMenu = $menuFrame->Menubutton(-text => 'Colors', -underline => 0);
foreach my $color (@ApplColors) {
    $colorsMenu->command(-label => $color, 
                          -command => sub { $mainFrame->setPalette($color); },
                          -background => $color);
}
$colorsMenu->pack(-side => 'left');

#
# help menu
#
my $helpMenu = $menuFrame->Menubutton(-text => 'Help', -underline => 0);
$helpMenu->command(-label => 'How to play', -command => [\&help] );
$helpMenu->command(-label => 'Run Demo', -command => [\&demo] );
$helpMenu->command(-label => 'About', -command => [\&help_about] );
$helpMenu->pack(-side => 'right');

#
# the actual table (game itself) ...
#
for( my $j=0; $j<$boardSizeY; $j++ ) {  # make each row a different frame
    $row[$j]=$board->Frame();
}

for( my $i=0; $i<$boardSizeX; $i++ ) {
    for( my $j=0; $j<$boardSizeY; $j++ ) {
        my $position=$i+$j*$boardSizeX;
        $square[$position]=$row[$j]
            ->Button(-image => $currentSet[0],
                     -padx => '0',
                     -pady => '0'  );

        $square[$position]->bind('<1>' => [\&square_click, $position] );
        $square[$position]->bind('<Leave>' => [\&square_leave, $position] );
        $square[$position]
            ->bind('<ButtonRelease>' => [\&square_leave, $position] );
        
        $square[$position]->pack(-side => 'left');
    }
}

for( my $j=0; $j<$boardSizeY; $j++ ) {
    $row[$j]->pack( -side => 'top');
}

#
# if the higth of the screen is less than about 1000 pixels, the window
# in "portrait" layout will be higher than a screen itself. to avoid this
# situation where the window does not fit the screen (user can't see the 
# whole thing...) i'll activate the "landscape" option if the available 
# screen higth is not enough to fit the window. 
#
# both "landscape" and "portrait" layout shall still be available from 
# the menu. i could disable this for smaller screen size, but i find that 
# would be confusing confusing. 
#

if( $mainFrame->screenheight() < 1000 ) {
    if($verbose) { print "Screen height = ".$mainFrame->screenheight()
                         ." < 1000  - use 'landscape' layout\n"; }
    $landscape = 1;
}

&re_pack;
&pick_random;
&add_kugel;

if( $] < 5.005 ) { 
    show_error( "\nThis program has been tested \n".
                "with Perl version 5.00502\n\n".
                "You might have problems trying\n".
                "to use it with older version\n\n".
                "It might be healthy to upgrade\n".
                "your Perl installation anyway."
              )
}
if( $tkVersion < 800.012 ) {
    show_error( "\n\n\n".
                "This program has been tested\n".
                "with Tk module release 800.012\n".
                "You might have problems trying\n".
                "to use it with older version\n"
              )
}

if( ! -w $scoreFile ) {
    show_error( "\nFile: $scoreFile\n".
                "not writable for user ID $<\n\n".
                "Change permissions for the \n".
                "scorefile, if you wish to \n".
                "be able to enter the high \n".
                "score list "
              )
}

MainLoop;

sub square_click {
    # bind will pass the Tk object as a first 
    # argument and i don't need it.
    shift;
    my $position = shift;
    if( defined( $selected )) {
        if( $square[$position]->cget(-image) == $currentSet[0]  &&  # move (?)
            path_exists($selected,$position)) {
            $square[$position]->configure(
                -image => $square[$selected]->cget(-image),
                -relief => 'raised' );
            $square[$selected]->configure(-image => $currentSet[0],
                                          -relief => 'raised' );
            $selected=undef;
            if( my $numDeleted = &delete_kugel($position) ) {
                $score += $numDeleted;
                $scoreLabel->configure(
                    -text => "Score: ".(sprintf '%04d',$score));
                $filled -= $numDeleted;

                $numFilledLabel->configure(
                    -text => (sprintf '%02d',
                             int($filled/($boardSizeX*$boardSizeY)*100))
                             ."\% Filled");
                #
                # empty label no longer used, but it might be in the future
                #
                $numEmptyLabel->configure(
                    -text => "Empty: ".(sprintf '%02d',
                                       ($boardSizeX*$boardSizeY-$filled))
                    ." (".(sprintf '%02d',
                           int(100*($boardSizeX*$boardSizeY-$filled)
                          /($boardSizeX*$boardSizeY)))."\%)");
                #
                # title of main window
                #
                $mainFrame->configure( -title => "Kugel Score: $score");
            }
            else {
                &add_kugel;
            }
        }
        else {
            if( $square[$position]->cget(-image) != $currentSet[0] ) {
                $square[$selected]->configure(-relief => 'raised' );
                $square[$position]->configure(-relief => 'sunken' );
                $selected=$position;  # new select
            }
        }
    }
    elsif( $square[$position]->cget(-image) != $currentSet[0]) {  # new select
        $square[$position]->configure(-relief => 'sunken' );
        $selected=$position;
    }
}

sub square_leave {
    # bind will pass the Tk object as a first
    # argument and i don't need it.
    shift;
    my $position = shift;

    # let the selected square stay 'suunken'
    if( defined($selected) && $selected==$position ) {
        $square[$position]->configure(-relief => 'sunken' );
    }
}

sub add_kugel {
    if( @topKugel + $filled >= $boardSizeX*$boardSizeY ) {  # no space left
        &game_over;
    }

    foreach my $kugel (@topKugel) {
        my $position ;
        do { 
            $position = int rand $boardSizeX*$boardSizeY; 
        } until ( $square[$position]->cget(-image) == $currentSet[0] );
        $square[$position]->configure(-image => $kugel->cget(-image) );
        if( my $numDeleted = &delete_kugel($position) ) {
            $filled -= $numDeleted;
            $numFilledLabel->configure(
                -text => (sprintf '%02d',
                         int($filled/($boardSizeX*$boardSizeY)*100))
                         ."\% Filled");
            $numEmptyLabel->configure(
                -text => "Empty: ".(sprintf '%02d',
                                   ($boardSizeX*$boardSizeY-$filled))." ("
                .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
                      /($boardSizeX*$boardSizeY)))."\%)");
        }

    }
    $filled += @topKugel ;  # add to number of kugels displayed
    $numFilledLabel->configure(-text => (sprintf '%02d',
                                int($filled/($boardSizeX*$boardSizeY)*100))
                               ."\% Filled");
    $numEmptyLabel->configure(
        -text => "Empty: ".(sprintf '%02d',($boardSizeX*$boardSizeY-$filled))
        ." ("
        .(sprintf '%02d',int(100*($boardSizeX*$boardSizeY-$filled)
              /($boardSizeX*$boardSizeY)))."\%)");
    &pick_random;
}

sub pick_random {
    foreach my $kugel (@topKugel) {  # delete the old ones
        $kugel->packForget();
    }
    @topKugel = () ;  # i hope garbage collection will stop the leak
    my $kugelsPicked =0 ;
    while( $kugelsPicked < $addKugel ) {
        push @topKugel, 
             $nextFrame->Label(-image=>$currentSet[(int rand $setOf)+1]);
        $kugelsPicked++;
    }
    &pack_next_kugel;
}

sub delete_kugel {
    my $position = shift;

    my $returnScore =0 ;

    my %numberOfKugels;  # number of kugels on each side

    $numberOfKugels{left}        = 0 ; # this init might not be necessary
    $numberOfKugels{right}       = 0 ;
    $numberOfKugels{above}       = 0 ;
    $numberOfKugels{bellow}      = 0 ;
    $numberOfKugels{aboveleft}   = 0 ;
    $numberOfKugels{aboveright}  = 0 ;
    $numberOfKugels{bellowleft}  = 0 ;
    $numberOfKugels{bellowright} = 0 ;

    #
    # left
    #
    for( my $tempPos = $position -1 ;
         ($tempPos+1)%$boardSizeX  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos--) {
        $numberOfKugels{left}++
    }

    #
    # right
    #
    for( my $tempPos = $position +1 ;
         $tempPos%$boardSizeX  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos++) {
        $numberOfKugels{right}++
    }

    #
    # above
    #
    for( my $tempPos = $position -$boardSizeX ;
         $tempPos>=0  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX ) {
        $numberOfKugels{above}++
    }

    #
    # bellow
    #
    for( my $tempPos = $position +$boardSizeX ;
         $tempPos < $boardSizeX*$boardSizeY  &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX ) {
        $numberOfKugels{bellow}++
    }

    #
    # aboveleft
    #
    for( my $tempPos = $position -$boardSizeX -1 ;
         $tempPos>=0  && ($tempPos+1)%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX+1 ) {
        $numberOfKugels{aboveleft}++
    }

    #
    # aboveright
    #
    for( my $tempPos = $position -$boardSizeX +1 ;
         $tempPos>=0  && $tempPos%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos -= $boardSizeX-1 ) {
        $numberOfKugels{aboveright}++
    }

    #
    # bellowleft
    #
    for( my $tempPos = $position +$boardSizeX-1 ;
         $tempPos < $boardSizeX*$boardSizeY  && ($tempPos+1)%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX-1 ) {
        $numberOfKugels{bellowleft}++
    }

    #
    # bellowright
    #
    for( my $tempPos = $position +$boardSizeX+1 ;
         $tempPos < $boardSizeX*$boardSizeY  && $tempPos%$boardSizeX &&
         $square[$tempPos]->cget(-image) == $square[$position]->cget(-image) ;
         $tempPos += $boardSizeX+1 ) {
        $numberOfKugels{bellowright}++
    }

    if( $numberOfKugels{right} + $numberOfKugels{left} >= $delKugel ) {
        for( my $count=1 ; $count<=$numberOfKugels{left}; $count++) {
            $square[$position-$count]->configure(-image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        for( my $count=1 ; $count<=$numberOfKugels{right}; $count++) {
            $square[$position+$count]->configure(-image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore += $numberOfKugels{right} + $numberOfKugels{left} ;
    }
    if( $numberOfKugels{above} + $numberOfKugels{bellow} >= $delKugel ) {
        for( my $count=1 ; $count<=$numberOfKugels{above}; $count++) {
            $square[$position-$count*$boardSizeX]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        for( my $count=1 ; $count<=$numberOfKugels{bellow}; $count++) {
            $square[$position+$count*$boardSizeX]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore += $numberOfKugels{above} + $numberOfKugels{bellow} ;
    }
    if( $numberOfKugels{aboveleft} + $numberOfKugels{bellowright} >=$delKugel){
        for( my $count=1 ; $count<=$numberOfKugels{aboveleft}; $count++) {
            $square[$position-$count*($boardSizeX+1)]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        for( my $count=1 ; $count<=$numberOfKugels{bellowright}; $count++) {
            $square[$position+$count*($boardSizeX+1)]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore +=$numberOfKugels{aboveleft}+$numberOfKugels{bellowright};
    }
    if( $numberOfKugels{aboveright} + $numberOfKugels{bellowleft} >=$delKugel){
        for( my $count=1 ; $count<=$numberOfKugels{aboveright}; $count++) {
            $square[$position-$count*($boardSizeX-1)]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        for( my $count=1 ; $count<=$numberOfKugels{bellowleft}; $count++) {
            $square[$position+$count*($boardSizeX-1)]->configure(
                                                 -image => $currentSet[0],
                                                 -relief => 'raised' );
        }
        $square[$position]->configure(-image => $currentSet[0],
                                      -relief => 'raised' );
        $returnScore +=$numberOfKugels{aboveright}+$numberOfKugels{bellowleft};
    }

    if( $returnScore == 0 ) { return $returnScore; }
    else { return $returnScore+1; }
}

sub kugel_choice {
    my $kugelChoiceDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                                -title => "Kugel Choice",
                                                -padx => '0');

    my %frame ;

    foreach my $dir (@dir) {
        $frame{$dir}=$kugelChoiceDialog->Frame(-relief =>'raised', 
                                               -borderwidth =>2);
        if( $#{$photo{$dir}} >= $setOf ) {
            foreach my $photo (@{$photo{$dir}}) {
                $frame{$dir}->Button(-image => $photo,
                                     -command => [\&new_kugel_choice, $dir] 
                              )->pack(-side => 'left', 
                                      -fill => 'x', 
                                      -expand => 'yes');
            }
        }
        else {
            foreach my $photo (@{$photo{$dir}}) {
                $frame{$dir}->Button(-image => $photo,
                                     -state => 'disabled',
                              )->pack(-side => 'left',
                                      -fill => 'x',
                                      -expand => 'yes');
            }
        }
    }

    foreach my $dir (@dir) { $frame{$dir}->pack(); }

    $kugelChoiceDialog->Show();
}

sub new_kugel_choice {
    my $choice=shift;
    my @newPhoto;

    foreach my $dir (@dir) {
        if( $choice eq $dir ) {
            @newPhoto = @{$photo{$dir}}
        }
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position = $i+$j*$boardSizeX;
            for( my $p=0; $p<=$#currentSet; $p++) {
                if( $square[$position]->cget(-image) == $currentSet[$p] ) {
                    if( defined $newPhoto[$p] ) {
                        $square[$position]->configure(-image => $newPhoto[$p]);
                    }
                    else { 
                        # only coding error can bring me here
                        if($verbose) { print "not enough kugels in new set\n";}
                        @currentSet=@newPhoto;
                        $setOf = $#currentSet;
                        if($verbose) { print "setOf = $setOf\n"; }
                        &full_restart;  # not enough kugels in new set
                    }
                }
            }
        }
    }
    for( my $i=0; $i<$addKugel ; $i++ ){
        for( my $p=0; $p<=$#currentSet; $p++) {
            if( $topKugel[$i]->cget(-image) == $currentSet[$p] ) {
                if( defined $newPhoto[$p] ) {
                    $topKugel[$i]->configure(-image => $newPhoto[$p]);
                }
                else {
                    # only coding error can bring me here
                    if($verbose) { print "not enough kugels in new set (2)\n"; }
                    @currentSet=@newPhoto;
                    $setOf = $#currentSet;
                    if($verbose) { print "setOf = $setOf\n"; }
                    &full_restart;  # not enough kugels in new set
                }
            }
        }
    }
    @currentSet=@newPhoto;
}

sub help_about {
    my $helpAboutDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                              -title => "Help About",
                                              -padx => '0');
    my $frame1=$helpAboutDialog->Frame(-relief =>'flat', -borderwidth =>2);

    my $text1 = $frame1->Text(-relief => 'flat',
                              -width => 42, 
                              -height => 10,
                              -cursor => 'top_left_arrow',
                              -font => $helpFont,
                              -setgrid => 'yes' );
    $text1->insert('end',"\n This program was made by Vlatko Primorac\n");
    $text1->insert('end',"\n vlp\@geocities.com\n");
    $text1->insert('end',"\n You may freely distribute it under \n");
    $text1->insert('end',"\n GNU general public license\n");

    $text1->pack(-side => 'left', -fill => 'x', -expand => 'yes');

    $frame1->pack();
    $helpAboutDialog->Show();

}

sub game_over {

    if( !defined($scores[9]) || $score>substr($scores[9],0,4) ) {
        &write_score;
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0],
                                          -relief => 'raised',     );
        }
    }
    $score=0;
    $didSkip=0;
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $filled=0;
    &pick_random;
}

sub game_restart {
    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0],
                                          -relief => 'raised',     );
        }
    }
    $score=0;
    $didSkip=0;
    $filled=0;
    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    &pick_random;

    # there is a remote possibility this will result in an
    # empty board, but that's likely only on "extreme" sets
    # like delete by 2 and set of 1. in that case, the game
    # should be re-restarted.
    while( $filled==0 ) {
        if($verbose) { print "populating the board, calling add_kugel\n"; }
        &add_kugel;
    }
    $mainFrame->configure( -title => "Kugel");
}

sub full_restart {

    $board->packForget();
    foreach my $row    (@row)    { $row->packForget(); }
    foreach my $square (@square) { $square->packForget(); }

    for( my $j=0; $j<$boardSizeY; $j++ ) {  # make each row a different frame
        $row[$j]=$board->Frame();
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]=$row[$j]
                ->Button(-image => $currentSet[0],
                         -padx => '0',
                         -pady => '0',
                         -relief => 'raised',
                         # -command => [\&square_click, $position]
            );

            # "bind" here 
            $square[$position]->bind('<1>' => [\&square_click, $position]);
            $square[$position]->bind('<Leave>' => [\&square_leave, $position]);
            $square[$position]
                ->bind('<ButtonRelease>' => [\&square_leave, $position]);

            $square[$position]->pack(-side => 'left');
        }
    }

    for( my $j=0; $j<$boardSizeY; $j++ ) {
        $row[$j]->pack( -side => 'top');
    }

    for( my $i=0; $i<$boardSizeX; $i++ ) {          # set the board blank
        for( my $j=0; $j<$boardSizeY; $j++ ) {
            my $position=$i+$j*$boardSizeX;
            $square[$position]->configure(-image => $currentSet[0]);
        }
    }
    $score=0;
    $didSkip=0;
    $filled=0;

    $scoreLabel->configure( -text => "Score: ".(sprintf '%04d',$score));
    $setOfLabel->configure( -text => "Set Of $setOf ");
    $addLabel  ->configure( -text => "Add by $addKugel ");
    $delLabel  ->configure( -text => "Delete by ".($delKugel+1) );
    # $numFilledLabel->configure();

    &pick_random;

    # there is a remote possibility this will result in an
    # empty board, but that's likely only on "extreme" sets
    # like delete by 2 and set of 1. in that case, the game 
    # should be re-restarted.
    while( $filled==0 ) {
        if($verbose) { print "populating the board, calling add_kugel\n"; }
        &add_kugel;
    }
    $board->pack();
    $mainFrame->configure( -title => "Kugel");
}

#
# determine if the clear path exists between two positions
#
sub path_exists {

    my $position1 = shift;
    my $position2 = shift;

    if($verbose) { print "sub path_exists($position1,$position2)\n"; }

    my @board;
    my @animate;  # squares to animate, if needed

    my $longest = $boardSizeX*$boardSizeY; # longest possible path (?)

    for( my $i=0; $i<$boardSizeX*$boardSizeY; $i++ ) { $board[$i]=$longest+1; }

    $board[$position1] =1 ;


    for( my $i=1; $i<$longest; $i++ ) {
        for( my $position=0; $position<$boardSizeX*$boardSizeY; $position++ ) {
            if( $board[$position] == $i ) {
                if($verbose) { print "position=$position reached\n"; }
                #
                # left
                #
                if( $position%$boardSizeX ) {
                    if( $position-1 == $position2 ) { goto bingo; }
                    if( $square[$position-1]->cget(-image) == $currentSet[0] &&
                        $board[$position-1] > $i+1 ) {
                        $board[$position-1] = $i+1;
                    }
                }
                #
                # right
                #
                if( ($position+1)%$boardSizeX ) {
                    if( $position+1 == $position2 ) { goto bingo; }
                    if( $square[$position+1]->cget(-image) == $currentSet[0] &&
                        $board[$position+1] > $i+1 ) {
                        $board[$position+1] = $i+1;
                    }
                }
                #
                # above
                #
                if( $position - $boardSizeX  >= 0 ) {
                    if( $position-$boardSizeX == $position2 ) { goto bingo; }
                    if( $square[$position-$boardSizeX]
                        ->cget(-image)==$currentSet[0]  &&
                        $board[$position-$boardSizeX] > $i+1 ) {
                        $board[$position-$boardSizeX] = $i+1;
                    }
                }
                #
                # bellow
                #
                if( $position +$boardSizeX < $boardSizeX*$boardSizeY ) {
                    if( $position+$boardSizeX == $position2 ) { goto bingo; }
                    if( $square[$position+$boardSizeX]
                        ->cget(-image)==$currentSet[0]  &&
                        $board[$position+$boardSizeX] > $i+1 ) {
                        $board[$position+$boardSizeX] = $i+1;
                    }
                }
            }
        }
    }
bingo:
    if( $position2%$boardSizeX &&                             # left
        $board[$position2-1] <= $longest) {
        if($verbose) { print "Path length = $board[$position2-1]\n"; }
        return $board[$position2-1];
    } 
    if( ($position2+1)%$boardSizeX &&                         # right
        $board[$position2+1] <= $longest) {
        if($verbose) { print "Path length = $board[$position2+1]\n"; }
        return $board[$position2+1];
    }
    if( $position2-$boardSizeX >=0 &&                          # above
        $board[$position2-$boardSizeX] <=$longest) { 
        if($verbose) { print "Path length = $board[$position2-$boardSizeX]\n"; }
        return $board[$position2-$boardSizeX];
    }
    if( $position2+$boardSizeX < $boardSizeX*$boardSizeY &&   # bellow
        $board[$position2+$boardSizeX] <= $longest) {
        if($verbose) { print "Path length = $board[$position2+$boardSizeX]\n"; }
        return $board[$position2+$boardSizeX];
    }

    if($verbose) { print "Path not found\n"; }

    if( $allowSkip ) { 
        if($verbose) { print "Skip allowed: allowSkip=$allowSkip\n"; }
        $didSkip=1;
        return 1; 
    }

    return # return undef (false)
}

sub set_dimensions {
    my $dimensionsDialog = $mainFrame->Dialog( -buttons => undef, 
                                               -title => "Set Dimensions",
                                               -padx => '0');

    my $oldSizeX = $boardSizeX;
    my $oldSizeY = $boardSizeY;
    my $oldAdd   = $addKugel;
    my $oldDel   = $delKugel;
    my $oldSetOf = $setOf;

    my $scaleYtext = $dimensionsDialog->Text(-relief => 'flat',
                                             -width => 5,
                                             -height => 1,
                                             -font => $bigFont,
                                             -cursor => 'top_left_arrow',
                                             -setgrid => 'yes');
    $scaleYtext->insert('end','Rows:');

    my $scaleY = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 4,
        -to     => 16,
        -length => 200,
        -tickinterval => 2,
        -command => sub { $boardSizeY=shift; }
    );
    $scaleY->set($boardSizeY);

    $scaleYtext->pack();
    $scaleY->pack();

    my $scaleXtext = $dimensionsDialog->Text(-relief => 'flat',
                                             -width => 8,
                                             -height => 1,
                                             -font => $bigFont,
                                             -cursor => 'top_left_arrow',
                                             -setgrid => 'yes');
    $scaleXtext->insert('end','Columns:');

    my $scaleX = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 4,
        -to     => 16,
        -length => 200,
        -tickinterval => 2,
        -command => sub { $boardSizeX=shift; }
    );
    $scaleX->set($boardSizeX);

    $scaleXtext->pack();
    $scaleX->pack();

    my $addText = $dimensionsDialog->Text(-relief => 'flat',
                                          -width => 7,
                                          -height => 1,
                                          -font => $bigFont,
                                          -cursor => 'top_left_arrow',
                                          -setgrid => 'yes');
    $addText->insert('end','Add by:');

    my $add = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 2,
        -to     => 6,
        -length => 200,
        -tickinterval => 1,
        -command => sub { $addKugel=shift; }
    );
    $add->set($addKugel);

    $addText->pack();
    $add->pack();

    my $delText = $dimensionsDialog->Text(-relief => 'flat',
                                          -width => 10,
                                          -height => 1,
                                          -font => $bigFont,
                                          -cursor => 'top_left_arrow',
                                          -setgrid => 'yes');
    $delText->insert('end','Delete by:');

    my $del = $dimensionsDialog->Scale( 
        -orient => 'horizontal',
        -from   => 2,
        -to     => 6,
        -length => 200,
        -tickinterval => 1,
        -command => sub { $delKugel=(shift)-1; });
    $del->set($delKugel+1);

    $delText->pack();
    $del->pack();

    my $setOfText = $dimensionsDialog->Text(-relief => 'flat',
                                            -width => 10,
                                            -height => 1,
                                            -font => $bigFont,
                                            -cursor => 'top_left_arrow',
                                            -setgrid => 'yes');
    $setOfText->insert('end','Use set of:');

    my $set = $dimensionsDialog->Scale(
        -orient => 'horizontal',
        -from   => 1,
        -to     => $#currentSet,  # let it scale to the number available
        -length => 200,
        -tickinterval => 1,
        -command => sub { $setOf=shift; }
    );
    $set->set($setOf);

    $setOfText->pack();
    $set->pack();

    $dimensionsDialog->Show();

    if($verbose) { print "boardSizeX = $boardSizeX \n"; }
    if($verbose) { print "boardSizeY = $boardSizeY \n"; }
    if($verbose) { print "addKugel = $addKugel \n"; }
    if($verbose) { print "delKugel = $delKugel \n"; }
    if($verbose) { print "setOf = $setOf \n"; }

    if( $boardSizeX != $oldSizeX || $boardSizeY != $oldSizeY ||
        $oldAdd != $addKugel     || $oldDel != $delKugel     ||
        $oldSetOf != $setOf                                     ) {
        &full_restart;
    }

    &get_scores;  # get the scores for the new settings
}

sub get_scores {
    my $SCORES;       # file handler
    open(SCORES, $scoreFile ) || die "could not open file $scoreFile: $!"  ;

    @scores=();

    while( defined( my $line = <SCORES> ) ) {
        if( $line !~ /^Rows:\s+/ ) { next; }  # skip junk...
        if( $line =~ /^Rows: (\d+) Cols: (\d+) Add: (\d+) Del: (\d+) Set: (\d+) Skip: ([0,1]) Score: (\d{4}) Name: (\w.*)/ ) {
            my ($rows, $cols, $add, $del, $set, $skip, $score, $name) = 
               ($1,    $2,    $3,   $4,   $5,   $6,    $7,     $8);
            if( $rows eq $boardSizeY &&
                $cols eq $boardSizeX &&
                $add  == $addKugel   &&
                $del  == $delKugel+1 &&
                $set  == $setOf      &&
                $skip == $allowSkip    ) {
                push @scores, "$score Name: $name";
            }
        }
    }

    close SCORES;

    # reversed order, sort lexically
    @scores = sort {$b cmp $a} @scores;

    if( $verbose ) {
        print "Scores: \n";
        foreach my $scoreline (@scores) { print "$scoreline\n"; }
    }

    if( defined( $scores[0] )) {
        $scores[0] =~ /^(\d{4})/;
        $highestScore = $1;
        $scores[0] =~ /Name: (\w.*)/;
        $champion = $1;
    }
    else {
        $highestScore = "0000";
        $champion = "1st time played";
    }

    if( defined($highScoreLabel) ) {  # rewrite the highest score
        $highScoreLabel->configure( 
            -text => "$champion: ".(sprintf '%04d',$highestScore) );
    }

    if( $verbose ) { print "highest score: $highestScore\n"; }
    if( $verbose ) { print "champion: $champion\n"; }

}

sub write_score {
    my $SCORES;                                 # file handler
    my $name="enter your name here...";         # user input name

    my $scoreDialog = $mainFrame->Dialog(
        -buttons => undef, 
        -title => "Congratulations!!! You have made it to the hall of fame.",
        -padx => '0');

    my $frame1=$scoreDialog->Frame(-relief =>'raised', -borderwidth =>2);
    my $frame2=$scoreDialog->Frame(-relief =>'raised', -borderwidth =>2);

    my $scoreText = $scoreDialog->Text(-relief => 'flat',
                                       -width => 40,
                                       -height => 5,
                                       -font => $scoreFont,
                                       -cursor => 'top_left_arrow',
                                       -setgrid => 'yes');

    $scoreText->insert('end', "\n           Congratulations !!!\n\n");
    $scoreText->insert('end', " You have made it to the hall of fame.\n");
    $scoreText->pack();

    my $entry = $frame2->LabEntry( -relief => 'sunken', 
                                   -width => 40,
                                   -textvariable => \$name);
    $entry->grid(-sticky => 'ew');
    $entry->focus();

    $frame1->pack();
    $frame2->pack();
    $scoreDialog->Show();

    $score = sprintf '%04d',$score;  # format to 0000

    if( $name eq "enter your name here..." ) { return; }

    open(SCORES, ">>$scoreFile" ) || warn "could not open file $scoreFile: $!";

    print SCORES "Rows: $boardSizeY ";
    print SCORES "Cols: $boardSizeX ";
    print SCORES "Add: $addKugel ";
    print SCORES "Del: ".($delKugel+1)." ";
    print SCORES "Set: ".($setOf)." ";
    print SCORES "Skip: $didSkip ";
    print SCORES "Score: $score Name: $name\n";
    close SCORES;

    if( $verbose) {
        print "Rows: $boardSizeY ";
        print "Cols: $boardSizeX ";
        print "Add: $addKugel ";
        print "Del: ".($delKugel+1)." ";
        print "Skip: $didSkip ";
        print "Score: $score Name: $name\n"; 
    }

    &get_scores;  # reload the scores
}

sub show_hall_of_fame {
    my $fameDialog = $mainFrame->Dialog( -buttons => ['Close'], 
                                         -title => "Hall of Fame",
                                         -padx => '0');

    my $fameText = $fameDialog->Text(-relief => 'flat',
                                     -width => 25,
                                     -height => 11,
                                     -font => $scoreFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $fameText->insert('end', "\n       Hall of Fame: \n\n");

    foreach my $scoreline (@scores) {
        $scoreline =~ /^(\d{4}) Name: (\w.*)/;
        $fameText->insert('end', " $1 $2 \n");
    }

    $fameText->pack();
    $fameDialog->Show();
}

#
# error dialog
#
sub show_error {
    my $text = shift;
    my $errorDialog = $mainFrame->Dialog( -buttons => ['Close'],
                                         -title => "Error Message",
                                         -padx => '0');

    my $errorText = $errorDialog->Text(-relief => 'flat',
                                       -width => 30,
                                       -height => 10,
                                       -font => $scoreFont,
                                       -cursor => 'top_left_arrow',
                                       -setgrid => 'yes');

    $errorText->insert('end', $text);

    $errorText->pack();
    $errorDialog->Show();
}

sub help {
    my $helpDialog = $mainFrame->Dialog( -buttons => ['Close'],
                                         -title => "How to play the game",
                                         -padx => '0');

    my $helpText = $helpDialog->Text(-relief => 'flat',
                                     -width => 62,
                                     -height => 23,
                                     -font => $helpFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $helpText->insert('end', 
        "\n                       How to play it? \n\n");
    $helpText->insert('end',  "\n");
    $helpText->insert('end',  
        'This is version 2.0 of the perl/tk based  game I call "kugel",'."\n");
    $helpText->insert('end',  
        'you can  call it whatever you  please.  The game  is played by'."\n");
    $helpText->insert('end',  
        'moving  objects  on  the board  and putting  them  into  rows,'."\n");
    $helpText->insert('end',  
        'columns, diagonals.  Each time you create  a line of  the same'."\n");
    $helpText->insert('end',  
        'objects  ("kugels", balls...), the score  will  increment with'."\n");
    $helpText->insert('end',  
        'the  number  of objects  you  accomplished to line up.  In the'."\n");
    $helpText->insert('end',  
        'same time,  the lined up "kugels"  will  disappear, so you can'."\n");
    $helpText->insert('end',  
        'continue the game with more maneuver  space on the board.  Any'."\n");
    $helpText->insert('end',  
        'time  you  fail to create  a line of "kugels",  new ones  will'."\n");
    $helpText->insert('end',  
        'appear on the board. Once the board is full, the game is over.'."\n");
    $helpText->insert('end',  "\n");
    $helpText->insert('end',  
        "The latest and greatest version of the game is available at:\n");
    $helpText->insert('end',  
        "http://www.sunsite.unc.edu/pub/Linux/games/strategy/\n");
    $helpText->insert('end', "\n");
    $helpText->insert('end',  
        "Have fun. :)\n\nVlatko Primorac, 1998, 1999\nvlp\@geocities.com\n\n");

    $helpText->pack();

    $helpDialog->Show();

}

#
# run demo dialog
#
sub demo {
    my $demoDialog = $mainFrame->Dialog( -default_button => 'Start',
                                         -buttons => ['Done'],
                                         -title => "Game Demo",
                                         -padx => '0');

    my $fameText = $demoDialog->Text(-relief => 'flat',
                                     -width => 25,
                                     -height => 3,
                                     -font => $scoreFont,
                                     -cursor => 'top_left_arrow',
                                     -setgrid => 'yes');

    $fameText->insert('end', "\n       Game Demo: \n\n");
    $fameText->pack();

    my $startButton;
    my $stepButton;
    my $moveButton;

    $startButton = $demoDialog->Button( 
        -text => "Start",
        -command=> sub{ 
                       $startButton->packForget(); 
                       $moveButton->pack(); 
                       &game_restart;
                   },
        -font => $scoreFont);

    $stepButton = $demoDialog->Button(-text => "Step",
                                      -command=> sub{ 
                                                     &pick_random;
                                                     &add_kugel;
                                                     $stepButton->packForget();
                                                     $moveButton->pack();
                                                 },
                                      -font => $scoreFont);

    $moveButton = $demoDialog->Button(-text => "Move",
                                      -command=> sub{  
                                                     &auto_move;
                                                     $moveButton->packForget();
                                                     $stepButton->pack();
                                                 },
                                      -font => $scoreFont);
    $startButton->pack();

    $demoDialog->Show();

}

#
# calculate and make a move
#
sub auto_move {
    # righ now, i have no idea how to make
    # the best move, so i'll just pick one

    my $start; 
    my $destination;

    do{ $start = int rand ($boardSizeX*$boardSizeX) }
    until($square[$start]->cget(-image) != $currentSet[0]);

    do{ $destination = int rand ($boardSizeX*$boardSizeX) }
    until($square[$destination]->cget(-image) == $currentSet[0]);

    if( $verbose) { print "Start: $start, Destination: $destination\n" };

    $square[$destination]->configure( -image => $square[$start]->cget(-image));
        # -image $square[$start]->configure(cget(-image));

    $square[$start]->configure(-image => $currentSet[0]);
}

# 
# re-pack everything
#
sub re_pack {

    $board->packForget();               # packForget all
    $nextFrame->packForget();
    $menuFrame->packForget();
    $semaphoreFrame->packForget();
    $bottomFrame->packForget();
    
    $menuFrame->pack(-fill => 'x');     # and re-pack all

    if( $showInfo ) {
        $scoreFrame->packForget();
        $infoFrame->packForget();
        $highScoreLabel->packForget();
        $scoreLabel->packForget();

        if( $landscape ) {
            $highScoreLabel->pack(-side =>'top', -expand =>'yes', -fill =>'x');
            $scoreLabel    ->pack(-side =>'top', -expand =>'yes', -fill =>'x');
        } else {
            $highScoreLabel->pack(-side =>'left', -expand =>'yes',-fill =>'x');
            $scoreLabel    ->pack(-side =>'left', -expand =>'yes',-fill =>'x');
        }

        if( $landscape ) {
            $setOfLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $addLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $delLabel->pack(-side => 'top', -expand => 'yes', -fill => 'x');
            $numFilledLabel->pack(-side =>'top', -expand =>'yes', -fill =>'x');
        } else {
            $setOfLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $addLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $delLabel->pack(-side => 'left', -expand => 'yes', -fill => 'x');
            $numFilledLabel->pack(-side =>'left', -expand =>'yes',-fill =>'x');
        }

        $scoreFrame->pack( -side => 'top', -fill => 'x', -expand => 'yes');
        $infoFrame->pack( -side => 'bottom', -fill => 'x', -expand => 'yes');

        if( $landscape ) {
            $semaphoreFrame->pack( -side => 'right', 
                                   -fill => 'y', 
                                   -expand => 'yes');
        } else {
            $semaphoreFrame->pack( -side => 'top', 
                                   -fill => 'x', 
                                   -expand => 'yes');
        }
    }

    if( $showNext ) {
        &pack_next_kugel;
        if( $landscape ) {
            $nextFrame->pack( -side => 'left',  -expand => 'no');
        } else {
            $nextFrame->pack( -side => 'top',  -expand => 'no');
        }
    }

    if( $showShortcuts ) { 
        $bottomFrame->pack( -side => 'bottom');
    }

    $board->pack();

    if($verbose) {
        print "Screen width:  ".$mainFrame->screenwidth()."\n";
        print "Screen height: ".$mainFrame->screenheight()."\n";
        print "Window width:  ".$mainFrame->width()." (to be updated)\n";
        print "Window height: ".$mainFrame->height()." (to be updated)\n";
    }
}

#
# pack the "next" kugel set
#
sub pack_next_kugel {
    foreach my $kugel (@topKugel) { $kugel->packForget() }
    unless( $showNext ) { return }
    foreach my $kugel (@topKugel) {
        if( $landscape ) {
            $kugel->pack(-side => 'top', -fill => 'x', -expand => 'yes');
        } else {
            $kugel->pack(-side => 'left', -fill => 'x', -expand => 'yes');
        }    
    }
}
